-----------------------------------------------------------------------------
-- |
-- Module      :  RefacSlicing
-- Copyright   :  (c) Christopher Brown 2005
--
-- Maintainer  :  cmb21@kent.ac.uk
-- Stability   :  provisional
-- Portability :  portable
--
-- This module contains a transformation for HaRe.
-- Symoblic Evaluation on tuples.
-- creates functions which evaluate tha expressions
-- within the return value of a function.
--  e.g.
--
-- @ f x y = (x, y) @
--
-- @ f1 x = x @
--
-- @ f2 y = y @
--
-----------------------------------------------------------------------------

module RefacSlicing where
 
import AbstractIO
import Maybe
import List
import RefacUtils
import RefacRedunDec

data Patt = Match HsMatchP | MyPat HsDeclP | Def [Char]

refacSlicing args
  = do let 
           fileName    = args!!0
           beginRow   = read (args!!1)::Int
           beginCol   = read (args!!2)::Int
           endRow     = read (args!!3)::Int
           endCol     = read (args!!4)::Int
       AbstractIO.putStrLn "refacTuple"

       -- Parse the input file.
       modInfo@(inscps, exps, mod, tokList) <- parseSourceFile fileName

       -- Find the function that's been highlighted as the refactree

       let (loc, pnt, pats, exp, wh, p)
             = findDefNameAndExp tokList
                                 (beginRow, beginCol)
                                 (endRow, endCol)
                                 mod 

       let newExp = locToExp (beginRow, beginCol) (endRow, endCol) tokList mod
       AbstractIO.putStrLn ((show newExp) ++ "\n")
       if newExp == defaultExp  
            then do
               error "Program slicing can only be performed on an expression."
            else do

               res <- checkLetAndLambda (beginRow, beginCol) (endRow, endCol) tokList wh newExp exp
               let res' = correctExpression res exp
               newExp' <- checkIfOrLet res' newExp tokList (beginRow, beginCol) (endRow, endCol) 
               (refactoredExp,wh') <- checkCase newExp' newExp wh
               refactoredMod <- applyRefac (sliceSubExp p exp refactoredExp wh' (head loc) pnt pats) (Just modInfo) fileName
               writeRefactoredFiles False [refactoredMod]
               AbstractIO.putStrLn "Completed.\n"
                where
                   correctExpression [] exp  = exp
                   correctExpression (x:xs) exp
                                   | x == defaultExp = correctExpression xs exp
                                   | otherwise = x

checkCase (Exp(HsParen e)) high wh 
                = do 
                     res <- checkCase e high wh
                     return res
checkCase (Exp(HsCase e alts)) high wh
                = do
                     newWhere <- checkAlts alts
                     return (high, (newWhere++wh))
                        where
                           checkAlts [] = do
                                             return []
                           checkAlts ((HsAlt _ _ _ ds):xs) 
                              = do
                                   (decls, newExp) <- removeRedun ds high
                                   res <- checkAlts xs
                                   return (decls ++ res)

checkCase (exp@e) high wh  = do
                              decls <- removeRedunWhere wh e
                              return (exp, decls)


checkIfOrLet (Exp(HsParen e)) newExp tokList start end
                = do
                     res <- checkIfOrLet e newExp tokList start end
                     return (Exp(HsParen res))

checkIfOrLet (Exp(HsIf e1 e2 e3)) newExp tokList start end
                = do
                     let expRes = findExpInIf tokList
                                              start
                                              end
                                              e2
                     if expRes == defaultExp
                         then do
                                 let expRes' = findExpInIf tokList
                                                           start
                                                           end
                                                           e3
                                 if expRes' == defaultExp 
                                      then return (Exp(HsIf e1 e2 e3)) 
                                      else return (newExp)
                         else return (newExp)
checkIfOrLet (Exp(HsLet ds e)) newExp _ _ _ 
                = do
                     return (Exp(HsLet ds newExp))

checkIfOrLet (exp@_) newExp _ _ _ = do
                               return newExp

sliceSubExp p old exp wh loc pnt pats (_,_, mod) = do
                                                 -- find all the free variable that are in 
                                                 -- the highlighted subexpression 
                                                 --(free, dec) <- hsFreeAndDeclaredPNs exp
                                                 --let decls = checkFreeInWhere free wh
                                                 (decls, newExp) <- removeRedun wh exp
                                                 
                                                 -- let newPNT = changeName newName pnt
                                                 mod' <- updating p mod loc pnt pats newExp decls
                                                 --mod' <- addDecl mod Nothing ((newDecl loc newPNT pats newExp decls), Nothing) True
                                                 return mod'
                                               
changeName newName (PNT (PN (UnQual _) (G modName _ optSrc)) Value s) 
                  = PNT (PN (UnQual newName) (G modName newName optSrc)) Value s
                  
updating (match@(Match x)) mod loc pnt pats rhs ds = do 
                                                       mod' <- update x (newMatch loc pnt pats rhs ds) mod
                                                       return mod'
updating (pat@(MyPat x)) mod loc pnt pats rhs ds = do 
                                                      mod' <- update x (newDecl loc pnt pats rhs ds) mod
                                                      return mod'

newMatch loc pnt pats rhs ds = HsMatch loc pnt pats (HsBody rhs) ds 
newDecl loc pnt pats rhs ds = Dec (HsFunBind loc [HsMatch loc pnt pats (HsBody rhs) ds] )

checkFreeInWhere :: [PName] -> [HsDeclP] -> [HsDeclP]
checkFreeInWhere [] _ = []
checkFreeInWhere _ [] = []
checkFreeInWhere (p:ps) list = (checkSinInWhere p list) ++ (checkFreeInWhere ps list)
                                 where
                                   checkSinInWhere :: PName -> [HsDeclP] -> [HsDeclP]
                                   checkSinInWhere p [] = []
                                   checkSinInWhere p (x:xs) 
                                      | defines p x = [x]
                                      | otherwise   = checkSinInWhere p xs



{-|
Takes the position of the highlighted code and returns
the function name, the list of arguments, the expression that has been
highlighted by the user, and any where\/let clauses associated with the
function. 
-}

{-findDefNameAndExp :: Term t => [PosToken] -- ^ The token stream for the 
                                          -- file to be
                                          -- refactored.
                  -> (Int, Int) -- ^ The beginning position of the highlighting.
                  -> (Int, Int) -- ^ The end position of the highlighting.
                  -> t          -- ^ The abstract syntax tree.
                  -> (SrcLoc, PNT, FunctionPats, HsExpP, WhereDecls) -- ^ A tuple of,
                     -- (the function name, the list of arguments,
                     -- the expression highlighted, any where\/let clauses
                     -- associated with the function).
 -}                    
findDefNameAndExp toks beginPos endPos t
  = fromMaybe ([], defaultPNT, [], defaultExp, [], Def [])
              (applyTU (once_tdTU (failTU `adhocTU` inMatch `adhocTU` inPat)) t)

    where
      --The selected sub-expression is in the rhs of a match
      inMatch (match@(HsMatch loc1  pnt pats (rhs@(HsBody e)) ds)::HsMatchP)
        | locToExp beginPos endPos toks rhs /= defaultExp
          = Just ([loc1], pnt, pats, e, ds, (Match match))
      inMatch _ = Nothing

      --The selected sub-expression is in the rhs of a pattern-binding
      inPat (pat@(Dec (HsPatBind loc1 ps (rhs@(HsBody e)) ds))::HsDeclP)
        | locToExp beginPos endPos toks rhs /= defaultExp
          = if isSimplePatBind pat
              then Just ([loc1], patToPNT ps, [], e, ds, (MyPat pat))
              else error "A complex pattern binding can not be generalised!"
      inPat _ = Nothing

findExpInIf toks beginPos endPos t
  = fromMaybe (defaultExp)
              (applyTU (once_tdTU (failTU `adhocTU` inExp)) t)
        where
          inExp (exp@(_)::HsExpP)
           | locToExp beginPos endPos toks exp /= defaultExp
              = Just (exp)
          inExp _ = Nothing

-- check whether the identifiers within the highlighted subexpression are 
-- used within the let statement - if they are then return the statement
checkLetAndLambda begin end tokList wh high exp
  = applyTU (full_tdTU (constTU [] `adhocTU` inLet)) exp
     where
           inLet (ex@(Exp(HsLet ds e))::HsExpP)
               = do 
                    let newExp = (Exp(HsLet ds high))
                    (decls, x) <- removeRedun wh newExp
                    res <- inLet e
                    if res == [defaultExp] then do
                                             return [x]
                                           else do
                                             let exp' = ammendDecs res x
                                             return [exp']
                                                where
                                                  ammendDecs ((Exp(HsLet ds e)):xs) ((Exp(HsLet decls e2))) 
                                                                 = (Exp(HsLet (decls++ds) e))
                                                  ammendDecs _ _ = defaultExp

           inLet (Exp(HsParen e)) = do res <- inLet e 
                                       return res
           inLet (Exp(HsInfixApp e1 _ e2)) = do
                                           res <- inLet e1
                                           if res == [defaultExp]
                                               then do
                                                  res' <- inLet e2
                                                  return res'
                                               else do
                                                  return res
           inLet (Exp(HsApp e1 e2)) = do
                                        res <- inLet e1
                                        if res == [defaultExp]
                                           then do
                                             res' <- inLet e2
                                             return res'
                                           else do
                                             return res
           inLet (Exp(HsCase e alts))
                = do 
                     return [(Exp(HsCase high alts))]

           inLet (exp@(Exp(HsDo _)))
               = do
                    if (locToExp begin end tokList exp) /= defaultExp
                       then do
                         error "Cannot perform a slicing on a monad"
                       else do return [exp] 

           inLet (exp@(Exp(HsListComp _))) = do
                                              if (locToExp begin end tokList exp) /= defaultExp
                                                then do
                                                  error "Cannot perform slicing with List Comprehension"
                                                else do return [defaultExp]

           inLet (_::HsExpP) = do
                                return [defaultExp]
                    

